#include "RecvFileThread.h"
#include "MySyncObject.h"		// CMySyncObjectを使うため
#include "SendRecvThread.h"		// CSendRecvThreadを使うため
#include "RingBuff.h"			// CRingBuffを使うため

extern void DispMenu();								// メニューの表示


CRecvFileThread::CRecvFileThread(CSendRecvThread *pCSendRecvThread, LPSTR pszFileName)
{
	// コンストラクタの引数を覚える
	m_pCSendRecvThread = pCSendRecvThread;
	m_pszFileName = (LPSTR)calloc(strlen(pszFileName) + 1, sizeof(char));
	DISABLE_C4996
	strcpy(m_pszFileName, pszFileName);
	ENABLE_C4996

	// 変数の初期化
	m_fIamZombie = FALSE;
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
	m_pCRingBuffCmd = new CRingBuff(CMD_BUFF_SIZE);	// ファイル受信要求/応答格納用
	m_iRecvFileStatus = RECV_FILE_STAT_IDLE;		// 遷移状態 RECV_FILE_STAT_IDLE/RECV_FILE_STAT_SIZE/RECV_FILE_STAT_START
													// RECV_FILE_STAT_ING/RECV_FILE_STAT_ABORT
	m_cRecvFileSizeRes = 0;							// ファイル受信 受信サイズ要求応答(0:初期値 1:OK -1:エラー)
	m_lRecvFileSize = 0;							// 受信サイズ (エラー時:0 存在時:実際のサイズ)
	m_cRecvFileAbortRes = 0;						// ファイル受信中断要求応答(0:初期値 1:OK -1:エラー)
	m_dwStart = 0;									// 応答待ちタイムアウトのため要求を行った時刻

	m_fdDstFile = -1;								// 受信ファイルディスクリプタ
}


CRecvFileThread::~CRecvFileThread()
{
	if (m_pCMySyncObject != NULL)
		m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	SAFE_FREE(m_pszFileName)
	SAFE_DELETE(m_pCRingBuffCmd)					// ファイル受信要求/応答格納用
}

UINT CRecvFileThread::DoWork()
{
	int			iRet = -1;

	while (!m_fStopFlag)
	{
		usleep(5);
		// ファイル受信遷移状態に応じた処理の実行
		// 応答に対する処理とファイル送信処理をこの関数で実行します
		iRet = DoRecvFileState();
		if ((iRet == 2) || (iRet == -1))	// 受信完了/エラー、このスレッドをゾンビにする
			break;
	}
	DispMenu();
	// ファイルのclose
	if (m_fdDstFile != -1)
		close(m_fdDstFile);
	m_pCMySyncObject->Lock();
	m_fIamZombie = TRUE;		// ファイル送信が完了/失敗でゾンビ
	m_pCMySyncObject->UnLock();
	return((iRet != -1) ? 0 : -1);
}

//----------------------------------------------
// function
//		ファイル受信遷移状態に応じた処理を実行
// parameter
//		なし
// return
//		0:何もしなかった 1:処理を実行した 2:送信完了 -1:エラー発生
//----------------------------------------------
int CRecvFileThread::DoRecvFileState()
{
	int iRet = 0;

	if (((iRet = AnalyzeCmd()) == -1) || (iRet == 2))	// 応答などの解析 返値が-1の時がエラー 2の時ファイル受信完了
		goto L_END;

	switch (m_iRecvFileStatus)
	{
		case RECV_FILE_STAT_IDLE:					// 初期状態->受信サイズ応答待ち
			if ((iRet = SendRecvFileSize()) == 1)	// 受信サイズ要求パケットを送信する
			{
				m_iRecvFileStatus = RECV_FILE_STAT_SIZE;
				m_dwStart = timeGetTime();
			}
			break;
		case RECV_FILE_STAT_SIZE:					// 受信サイズ応答待ち->ファイル受信開始要求
			if ((iRet = CheckRecvFileSizeRes(TIMEOUT_FOR_RES)) == 1)	// 応答結果チェック
			{
				if ((iRet = SendRecvFileStart()) == 1)	// ファイルをOpenして、受信開始要求パケットを送信する
				{
					m_iRecvFileStatus = RECV_FILE_STAT_START;
				}
			}
			break;
		case RECV_FILE_STAT_ABORT:				// ★中断完了待ち->終わり
			if ((iRet = CheckRecvFileAbortRes(TIMEOUT_FOR_RES)) == 2)
			{
				Locate(1, 10, 2);
				fprintf(stderr, "ファイル受信中断完了");
				m_iRecvFileStatus = RECV_FILE_STAT_IDLE;
			}
			break;
		default:
			break;
	}
L_END:
	if (iRet == -1)
	{
		Locate(1, 8, 1);
		fprintf(stderr, "ファイル受信エラー");
	}
	return(iRet);
}

//----------------------------------------------
// function
//		ファイルのサイズ要求パケットを送信する
// parameter
//		なし
// return
//		1/-1
//----------------------------------------------
int CRecvFileThread::SendRecvFileSize()
{
	int iRet = -1;

	m_cRecvFileSizeRes = 0;				// 応答受信準備
	// ファイル名の送信
	if (SendRecvFileSizePacket(m_pszFileName) == FALSE)
		goto L_END;
	iRet = 1;
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイルサイズ要求コマンドの送信
// parameter
//		LPSTR	pszFileName	[in]ファイル名
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendRecvFileSizePacket(LPSTR pszFileName)
{
	BOOL		fRet = FALSE;
	RecvFileSizeReqRec	*pRecvFileSizeReq = NULL;
	int			iPacketSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(RecvFileSizeReqRec);
	pRecvFileSizeReq = (RecvFileSizeReqRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileSizeReq->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileSizeReq->header.wCommand = htons(CMD_RECV_FILE_SIZE_REQ);
	pRecvFileSizeReq->header.wDataLen = htons(sizeof(FileInfoRec));

	Locate(1, 8, 2);
	fprintf(stderr, "ファイル受信 %s", pszFileName);
	// データ部にデータを格納(ファイル名)
	DISABLE_C4996
	strcpy(pRecvFileSizeReq->FileInfo.szFileName, pszFileName);
	ENABLE_C4996

	// ファイルサイズ要求パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileSizeReq, iPacketSize);

	SAFE_FREE(pRecvFileSizeReq);
	return(fRet);
}

//----------------------------------------------
// function
//		ファイル受信サイズ要求応答チェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 1:成功が来た -1:失敗またはタイムアウト 2:ファイルが存在しない
//----------------------------------------------
int	CRecvFileThread::CheckRecvFileSizeRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cRecvFileSizeRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cRecvFileSizeRes == 1)			// 成功応答
	{
		iRet = (m_lRecvFileSize == 0) ? 2 : 1; // ファイルサイズが0の時はファイルが存在しない
		Locate(1, 10, 2);
		fprintf(stderr, "ファイルが存在しない");
	}
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信開始要求
// parameter
//		なし
// return
//		1/-1
//----------------------------------------------
int CRecvFileThread::SendRecvFileStart()
{
	int		iRet = -1;
	long	lRecvFileSize;

	// ファイルのOpen
	DISABLE_C4996
	m_fdDstFile = open(m_pszFileName, O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	ENABLE_C4996

	// ファイル長0で要求すると、ファイル受信を実施しません
	lRecvFileSize = (m_fdDstFile == -1) ? 0 : m_lRecvFileSize;

	// ファイル名, ファイル長の送信
	if (SendRecvFileStartPacket(m_pszFileName, lRecvFileSize) == FALSE)
		goto L_END;
	iRet = (m_fdDstFile == -1) ? -1 : 1;
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信開始コマンドの送信
// parameter
//		LPSTR	pszFileName	[in]ファイル名
//		int		iFileLength	[in]ファイル長/0:ファイル受信実施しない
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendRecvFileStartPacket(LPSTR pszFileName, int iFileLength)
{
	BOOL		fRet = FALSE;
	RecvFileStartReqRec	*pRecvFileStartReq = NULL;
	int			iPacketSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(RecvFileStartReqRec);
	pRecvFileStartReq = (RecvFileStartReqRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileStartReq->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileStartReq->header.wCommand = htons(CMD_RECV_FILE_START_REQ);
	pRecvFileStartReq->header.wDataLen = htons(sizeof(FileInfoRec));

	Locate(1, 8, 2);
	fprintf(stderr, "ファイル受信 %s %d(%u)", pszFileName, iFileLength, timeGetTime());
	// データ部にデータを格納(ファイルサイズ、ファイル名)
	DISABLE_C4996
	strcpy(pRecvFileStartReq->FileInfo.szFileName, pszFileName);
	sprintf(pRecvFileStartReq->FileInfo.szFileSize, "%d", iFileLength);
	ENABLE_C4996

	// ファイル受信開始要求パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileStartReq, iPacketSize);

	SAFE_FREE(pRecvFileStartReq);
	return(fRet);
}

//----------------------------------------------
// function
//		受信中断
//		遷移状態変更、中断パケット送信
// parameter
//		なし
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::RecvFileAbort()
{
	BOOL		fRet = FALSE;

	// ファイル受信中の時だけ中断
	if (m_iRecvFileStatus != RECV_FILE_STAT_ING)
		goto L_END;

	m_iRecvFileStatus = RECV_FILE_STAT_ABORT;
	m_dwStart = timeGetTime();
	fRet = SendRecvFileAbortPacket();
L_END:
	return(fRet);
}

//----------------------------------------------
// function
//		ファイル受信の中断要求送信
// parameter
//		なし
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendRecvFileAbortPacket()
{
	BOOL		fRet = FALSE;
	RecvFileAbortReqRec	*pRecvFileAbortReq = NULL;
	int			iPacketSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(RecvFileAbortReqRec);
	pRecvFileAbortReq = (RecvFileAbortReqRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileAbortReq->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileAbortReq->header.wCommand = htons(CMD_RECV_FILE_ABORT_REQ);
	pRecvFileAbortReq->header.wDataLen = htons(0);

	Locate(1, 10, 2);
	fprintf(stderr, "ファイル受信中断要求送信");

	// ファイル受信中断パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileAbortReq, iPacketSize);
	SAFE_FREE(pRecvFileAbortReq);
	return(fRet);
}

//----------------------------------------------
// function
//		ファイル受信の中断要求応答チェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 2:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int	CRecvFileThread::CheckRecvFileAbortRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cRecvFileAbortRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cRecvFileAbortRes == 1)			// 成功応答
		iRet = 2;
L_END:
	if (iRet != 0)
	{
		if (m_fdDstFile != -1)
		{
			close(m_fdDstFile);
			unlink(m_pszFileName);
		}
		m_fdDstFile = -1;
	}
	return(iRet);
}


//----------------------------------------------
// function
//		ファイル受信コマンド/応答の解析
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗 2:ファイル受信完了
//----------------------------------------------
int CRecvFileThread::AnalyzeCmd()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	int			iSize = m_pCRingBuffCmd->GetReadableSize();
	int			iPacketSize, iDataSize;
	LPBYTE		pbPacket = NULL;

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする(パケット丸ごと格納されているはず)
	m_pCRingBuffCmd->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	wCmd = ntohs(Header.wCommand);
	iDataSize = ntohs(Header.wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);
	if (iSize < iPacketSize)				// パケットサイズに満たないときは何もしない
		goto L_END;

	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を読む
	m_pCRingBuffCmd->Read(pbPacket, iPacketSize);

	switch (wCmd)
	{
		case CMD_RECV_FILE_SIZE_RES:			// 受信サイズ要求応答受信
			iRet = RecvRecvFileSizeResPacket((RecvFileSizeResRec *)pbPacket);
			break;
		case CMD_RECV_FILE_ING:					// ファイル受信中(ファイル受信開始要求/ファイル受信中->ファイル受信中)
			iRet = RecvRecvFileIngPacket((RecvFileIngRec *)pbPacket);
			break;
		case CMD_RECV_FILE_END_REQ:				// ファイル受信完了要求受信(ファイル受信中->ファイル受信完了->初期状態)
			iRet = RecvRecvFileEndPacket((RecvFileEndReqRec *)pbPacket);
			break;
		case CMD_RECV_FILE_ABORT_RES:			// 受信中断要求応答受信(ファイル受信中->初期状態)
			iRet = RecvRecvFileAbortResPacket((RecvFileResRec *)pbPacket);
			break;
		default:
			iRet = -1;
			break;
	}
L_END:
	SAFE_FREE(pbPacket)
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信サイズ要求応答
// parameter
//		RecvFileSizeResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CRecvFileThread::RecvRecvFileSizeResPacket(RecvFileSizeResRec *pbPacket)
{
	int			iRet = 1;

	m_cRecvFileSizeRes = 1;
	m_lRecvFileSize = atol(pbPacket->FileInfo.szFileSize);
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信中
// parameter
//		RecvFileIngRec *pbPacket	[in]受信パケット
// return
//		0/1/-1
//----------------------------------------------
int CRecvFileThread::RecvRecvFileIngPacket(RecvFileIngRec *pbPacket)
{
	int			iRet = 0;
	long		lDataSize, lOffset;

	Locate(1, 9, 1);
	fprintf(stderr, "ファイル受信中 %s %s", pbPacket->Stat.szOffset, pbPacket->Stat.szSize);

	if (m_iRecvFileStatus == RECV_FILE_STAT_START)
		m_iRecvFileStatus = RECV_FILE_STAT_ING;
	if (m_iRecvFileStatus == RECV_FILE_STAT_ABORT)	//　中断処理中は書き込みを行わない
	{
		Locate(1, 10, 2);
		fprintf(stderr, "ファイル受信中断中");
		iRet = 1;
		goto L_END;
	}

	if (m_iRecvFileStatus != RECV_FILE_STAT_ING)
	{
		iRet = -1;				// 遷移状態エラーは終了
		goto L_END;
	}
	lDataSize = atol(pbPacket->Stat.szSize);
	lOffset = atol(pbPacket->Stat.szOffset);
	if (lDataSize > 0)
	{
		// ファイル操作失敗時エラーとして終了
		if (lseek(m_fdDstFile, lOffset, SEEK_SET) == -1)
		{
			iRet = -1;
			goto L_END;
		}
		if (write(m_fdDstFile, pbPacket->bData, lDataSize) != lDataSize)
		{
			iRet = -1;
			goto L_END;
		}
	}
	iRet = 1;
L_END:
	if (iRet == -1)				// エラー時オープンしているファイルがあればクローズ、削除
	{
		if (m_fdDstFile != -1)
		{
			close(m_fdDstFile);
			unlink(m_pszFileName);
		}
		m_fdDstFile = -1;
	}
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信完了
// parameter
//		RecvFileEndReqRec *pbPacket	[in]受信パケット
// return
//		-1/2
//----------------------------------------------
int CRecvFileThread::RecvRecvFileEndPacket(RecvFileEndReqRec *pbPacket)
{
	int			iRet = 0;
	long		lDataSize, lOffset;

	Locate(1, 10, 1);
	fprintf(stderr, "ファイル受信完了 %s %s(%u)", pbPacket->Stat.szOffset, pbPacket->Stat.szSize, timeGetTime());

	if (m_iRecvFileStatus != RECV_FILE_STAT_ING)
	{
		iRet = -1;				// 遷移状態エラーは終了
		goto L_END;
	}
	lDataSize = atol(pbPacket->Stat.szSize);
	lOffset = atol(pbPacket->Stat.szOffset);
	if (lDataSize > 0)
	{
		// ファイル操作失敗時エラーとして終了
		if (lseek(m_fdDstFile, lOffset, SEEK_SET) == -1)
		{
			iRet = -1;
			goto L_END;
		}
		if (write(m_fdDstFile, pbPacket->bData, lDataSize) != lDataSize)
		{
			iRet = -1;
			goto L_END;
		}
	}
	iRet = 2;
	SendRecvFileEndResPacket(1);		// 受信完了要求応答送信
L_END:
	if (m_fdDstFile != -1)
	{
		close(m_fdDstFile);
		if (iRet == -1)					// エラー時削除
			unlink(m_pszFileName);
	}
	m_iRecvFileStatus = RECV_FILE_STAT_IDLE;
	m_fdDstFile = -1;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信完了要求に対する応答を送信する
// parameter
//		char cResCode	[in]応答コード 1/-1
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendRecvFileEndResPacket(char cResCode)
{
	BOOL	fRet = FALSE;
	RecvFileResRec	*pRecvFileEndRes = NULL;
	int		iPacketSize;

	iPacketSize = sizeof(RecvFileResRec);
	pRecvFileEndRes = (RecvFileResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileEndRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileEndRes->header.wCommand = htons(CMD_RECV_FILE_END_RES);
	pRecvFileEndRes->header.wDataLen = htons(sizeof(char));

	pRecvFileEndRes->cResCode = cResCode;

	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileEndRes, iPacketSize);
	SAFE_FREE(pRecvFileEndRes)

	return(fRet);
}

//----------------------------------------------
// function
//		ファイル受信中断要求応答
// parameter
//		RecvFileResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CRecvFileThread::RecvRecvFileAbortResPacket(RecvFileResRec *pbPacket)
{
	int			iRet = 1;

	m_cRecvFileAbortRes = pbPacket->cResCode;
	return(iRet);
}



//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//	  TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::IsZombie()
{
	BOOL	fRet;

	m_pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pCMySyncObject->UnLock();
	return(fRet);
}
