#pragma once
#include "ThreadJob.h"
#include "define.h"					// パケット作成のため

#define	CMD_BUFF_SIZE	(1024 * 64)	// ファイル受信コマンド/応答格納リングバッファのサイズ

// ファイル受信遷移状態
#define RECV_FILE_STAT_IDLE		0	// 初期状態
#define RECV_FILE_STAT_SIZE		1	// ファイル受信 受信サイズ要求応答待ち
#define RECV_FILE_STAT_START	2	// ファイル受信開始要求
#define RECV_FILE_STAT_ING		3	// ファイル受信中
#define RECV_FILE_STAT_END		4	// ファイル受信完了受信
#define RECV_FILE_STAT_ABORT	5	// ファイル受信中断要求応答待ち


class CMySyncObject;				// CMySyncObjectを使うため
class CSendRecvThread;				// CSendRecvThreadを使うため
class CRingBuff;					// CRingBuffを使うため

class CRecvFileThread :
	public CThreadJob
{
public:
	CRecvFileThread(CSendRecvThread *pCSendRecvThread, LPSTR pszFileName);
	~CRecvFileThread();

	CRingBuff	*m_pCRingBuffCmd;			// ファイル受信コマンド/応答格納リングバッファ

	UINT DoWork() override;
	BOOL IsZombie();						// このスレッドはゾンビ状態か
	BOOL RecvFileAbort();					// ファイル送信の中断

private:
	LPSTR			m_pszFileName;			// 送信ファイル名
	CSendRecvThread	*m_pCSendRecvThread;	// 送受信スレッド(親スレッド)
	CMySyncObject	*m_pCMySyncObject;
	BOOL			m_fIamZombie;			// ゾンビ状態か(送信終了)

	int				m_iRecvFileStatus;		// 遷移状態 RECV_FILE_STAT_IDLE/RECV_FILE_STAT_SIZE/RECV_FILE_STAT_START
											// RECV_FILE_STAT_ING/RECV_FILE_STAT_ABORT

	char			m_cRecvFileSizeRes;		// ファイル受信 受信サイズ要求応答(0:初期値 1:OK -1:エラー)
	long			m_lRecvFileSize;		// 受信サイズ (エラー時:0 存在時:実際のサイズ)
	char			m_cRecvFileAbortRes;	// ファイル受信中断要求応答(0:初期値 1:OK -1:エラー)
	DWORD			m_dwStart;				// 応答待ちタイムアウトのため要求を行った時刻

	int				m_fdDstFile;			// 受信ファイルディスクリプタ

	int DoRecvFileState();					// 遷移状態に応じたファイル受信処理,応答処理を実行
	// ファイル受信処理
	int SendRecvFileSize();						// ファイル受信サイズ要求
	BOOL SendRecvFileSizePacket(LPSTR pszFileName);
	int	CheckRecvFileSizeRes(DWORD dwTimeOut);	// ファイル受信サイズ要求応答チェック
	int SendRecvFileStart();					// ファイル受信開始要求
	BOOL SendRecvFileStartPacket(LPSTR pszFileName, int iFileLength);
	BOOL SendRecvFileAbortPacket();				// ファイル受信の中断要求送信
	int	CheckRecvFileAbortRes(DWORD dwTimeOut);	// ファイル受信の中断要求応答チェック
	

	// 応答処理
	int AnalyzeCmd();							// ファイル受信コマンド/応答パケットの解析
	int RecvRecvFileSizeResPacket(RecvFileSizeResRec *pbPacket);	// ファイル受信サイズ要求応答
	int RecvRecvFileIngPacket(RecvFileIngRec *pbPacket);			// ファイル受信中
	int RecvRecvFileEndPacket(RecvFileEndReqRec *pbPacket);			// ファイル受信完了
	BOOL SendRecvFileEndResPacket(char cResCode);					// ファイル受信完了要求に対する応答を送信する
	int RecvRecvFileAbortResPacket(RecvFileResRec *pbPacket);		// ファイル受信中断要求応答
};

