#include "SendFileThread.h"
#include "MySyncObject.h"		// CMySyncObjectを使うため
#include "SendRecvThread.h"		// CSendRecvThreadを使うため
#include "RingBuff.h"			// CRingBuffを使うため

extern void DispMenu();								// メニューの表示

CSendFileThread::CSendFileThread(CSendRecvThread *pCSendRecvThread, LPSTR pszFileName)
{
	// コンストラクタの引数を覚える
	m_pCSendRecvThread = pCSendRecvThread;
	m_pszFileName = (LPSTR)calloc(strlen(pszFileName) + 1, sizeof(char));
	DISABLE_C4996
	strcpy(m_pszFileName, pszFileName);
	ENABLE_C4996

	// 変数の初期化
	m_fIamZombie = FALSE;
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
	m_pCRingBuffCmd = new CRingBuff(CMD_BUFF_SIZE);	// ファイル送信要求/応答格納用

	m_iSendFileStatus = SEND_FILE_STAT_IDLE;		// 遷移状態 SEND_FILE_STAT_IDLE/SEND_FILE_STAT_START/SEND_FILE_STAT_ING/SEND_FILE_STAT_END

	m_cSendFileStartRes = 0;						// ファイル送信開始要求応答(0:初期値 1:OK -1:エラー)
	m_cSendFileEndRes = 0;							// ファイル送信完了要求応答(0:初期値 1:OK -1:エラー)
	m_cSendFileAbortRes = 0;						// ファイル送信中断要求応答(0:初期値 1:OK -1:エラー)
	m_dwStart = 0;									// 要求を送信した時刻
	m_fdSrcFile = -1;								// 送信ファイルディスクリプタ
	m_lTotalSentSize = 0;							// 送信済みバイト数
	m_iWaitCount = 0;								// 送信待ちをした回数
}


CSendFileThread::~CSendFileThread()
{
	if(m_pCMySyncObject != NULL)
		m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	SAFE_FREE(m_pszFileName)
	SAFE_DELETE(m_pCRingBuffCmd)					// ファイル送信要求/応答格納用
}


UINT CSendFileThread::DoWork()
{
	int			iRet = -1;

	while (!m_fStopFlag)
	{
		usleep(5 * 1000);					// usleepを入れて他のスレッドにチャンスを与える
// ファイル送信遷移状態に応じた処理の実行
// 応答に対する処理とファイル送信処理をこの関数で実行しますF
		iRet = DoSendFileState();
		if ((iRet == 2) || (iRet == -1))	// 送信完了/エラー、このスレッドをゾンビにする
			break;
	}
	DispMenu();
	// ファイルのclose
	if (m_fdSrcFile != -1)
		close(m_fdSrcFile);
	m_pCMySyncObject->Lock();
	m_fIamZombie = TRUE;		// ファイル送信が完了/失敗でゾンビ
	m_pCMySyncObject->UnLock();
	return((iRet != -1) ? 0 : -1);
}

//----------------------------------------------
// function
//		ファイル送信遷移状態に応じた処理を実行
// parameter
//		なし
// return
//		0:何もしなかった 1:処理を実行した 2:送信完了 -1:エラー発生
//----------------------------------------------
int CSendFileThread::DoSendFileState()
{
	int iRet = 0;

	if (AnalyzeCmd() == -1)						// 応答などの解析 返値が-1の時がエラー
	{
		iRet = -1;
		goto L_END;
	}
	switch (m_iSendFileStatus)
	{
		case SEND_FILE_STAT_IDLE:				// 初期状態->送信開始応答待ち
			if ((iRet = SendFileStart()) == 1)	// ①ファイルをOpenして、送信開始要求パケットを送信する
			{
				m_iSendFileStatus = SEND_FILE_STAT_START;
				m_dwStart = timeGetTime();
			}
			break;
		case SEND_FILE_STAT_START:				// 送信開始応答待ち->送信中
			if ((iRet = CheckSendFileStartRes(TIMEOUT_FOR_RES)) == 1)	// ★②応答結果チェック
			{
				m_iSendFileStatus = SEND_FILE_STAT_ING;
				m_lTotalSentSize = 0;			// ファイル受信準備
				m_iWaitCount = 0;
			}
			break;
		case SEND_FILE_STAT_ING:				// 送信中->送信中/送信完了応答待ち
			if ((iRet = SendFileIng()) == 2)	// 送信中->送信完了応答待ち
			{
				iRet = 1;						// ★完了要求応答待ちまで完了としない
				m_iSendFileStatus = SEND_FILE_STAT_END;
				m_dwStart = timeGetTime();
			}
			break;
		case SEND_FILE_STAT_END:				// ★送信完了完応答待ち->終わり
			if ((iRet = CheckSendFileEndRes(TIMEOUT_FOR_RES)) == 2)
			{
				Locate(1, 10, 2);
				fprintf(stderr, "ファイル送信完了");
				m_iSendFileStatus = SEND_FILE_STAT_IDLE;
			}
			break;
		case SEND_FILE_STAT_ABORT:				// ★中断完了待ち->終わり
			if ((iRet = CheckSendFileAbortRes(TIMEOUT_FOR_RES)) == 2)
			{
				Locate(1, 10, 2);
				fprintf(stderr, "ファイル送信中断完了");
				m_iSendFileStatus = SEND_FILE_STAT_IDLE;
			}
			break;
		default:				// 知らない遷移状態
			iRet = -1;
			break;
	}
L_END:
	if (iRet == -1)
	{
		Locate(1, 8, 1);
		fprintf(stderr, "ファイル送信エラー");
	}
	return(iRet);
}

//----------------------------------------------
// function
//		①ファイルをOpenして、送信開始要求パケットを送信する
// parameter
//		なし
// return
//		1/-1
//----------------------------------------------
int CSendFileThread::SendFileStart()
{
	int iRet = -1;
	struct stat statInfo;			// ファイルの情報のための変数

	m_cSendFileStartRes = 0;		// 応答受信準備
	// ファイルのOpen
	DISABLE_C4996
	m_fdSrcFile = open(m_pszFileName, O_RDONLY);
	ENABLE_C4996
	if (m_fdSrcFile == -1)
		goto L_END;
	// ファイルの情報を取得(ファイルサイズ)
	if (fstat(m_fdSrcFile, &statInfo) != 0)
		goto L_END;
	m_lFileLength = statInfo.st_size;

	// ファイル名, ファイル長の送信
	if (SendSendFileStartPacket(m_pszFileName, m_lFileLength) == FALSE)
		goto L_END;
	iRet = 1;
L_END:
	return(iRet);
}


//----------------------------------------------
// function
//		①ファイル送信開始コマンドの送信
// parameter
//		LPSTR	pszFileName	[in]ファイル名
//		int		iFileLength	[in]ファイル長
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::SendSendFileStartPacket(LPSTR pszFileName, int iFileLength)
{
	BOOL		fRet = FALSE;
	SendFileStartReqRec	*pSendFileStartReq = NULL;
	int			iPacketSize;
	int			iSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(SendFileStartReqRec);
	pSendFileStartReq = (SendFileStartReqRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pSendFileStartReq->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileStartReq->header.wCommand = htons(CMD_SEND_FILE_START_REQ);
	pSendFileStartReq->header.wDataLen = htons(sizeof(FileInfoRec));

	Locate(1, 8, 2);
	fprintf(stderr, "ファイル送信 %s %d", pszFileName, iFileLength);

	DISABLE_C4996
	sprintf(pSendFileStartReq->FileInfo.szFileName, "%s", pszFileName);
	sprintf(pSendFileStartReq->FileInfo.szFileSize, "%d", iFileLength);
	ENABLE_C4996

	// ファイル送信開始要求パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileStartReq, iPacketSize);

	SAFE_FREE(pSendFileStartReq);
	return(fRet);
}


//----------------------------------------------
// function
//		②ファイル送信開始要求応答のチェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 1:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int CSendFileThread::CheckSendFileStartRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cSendFileStartRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cSendFileStartRes == 1)			// 成功応答
		iRet = 1;
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		⑤ファイル送信完了要求応答のチェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 2:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int CSendFileThread::CheckSendFileEndRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cSendFileEndRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cSendFileEndRes == 1)			// 成功応答
		iRet = 2;
L_END:
	return(iRet);
}


#define	READ_SIZE	(1024 * 10)			// ファイルの読み込み単位(ガバッと読みます)
//----------------------------------------------
// function
//		③④ファイルを読み込んでファイル送信中/ファイル送信完了パケットを送信する
//		この関数でエラーが起こったときは、中断コマンドを送るのも良いでしょう
// parameter
//		なし
// retun
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::SendFileIng()
{
	BOOL			fEnd = FALSE, fRet;
	LPBYTE			pbData = NULL;
	int				iRet = 0, iReadSize;
	SendFileIngRec	*pSendFileIng = NULL;
	int				iPacketSize;

	// ファイルを読み込む
	pbData = (LPBYTE)calloc(READ_SIZE, sizeof(BYTE));
	lseek(m_fdSrcFile, m_lTotalSentSize, SEEK_SET);
	iReadSize = read(m_fdSrcFile, pbData, READ_SIZE);
	iPacketSize = sizeof(HeaderRec) + sizeof(SendFileStatRec) + iReadSize;
	pSendFileIng = (SendFileIngRec *)calloc(iPacketSize, sizeof(BYTE));

	// 共通ヘッダ部にデータを格納する
	memcpy(pSendFileIng->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileIng->header.wDataLen = htons((WORD)sizeof(SendFileStatRec) + iReadSize);
	// データ部にデータを格納する
	DISABLE_C4996
	sprintf(pSendFileIng->Stat.szOffset, "%d", m_lTotalSentSize);	// 読み込み位置
	sprintf(pSendFileIng->Stat.szSize, "%d", iReadSize);			// 読み込んだデータ数
	ENABLE_C4996
	if (iReadSize > 0)				// ファイルの途中を送信
	{
		Locate(1, 9, 1);
		fprintf(stderr, "ファイル送信中 %d %d %d", m_lTotalSentSize, iReadSize, m_lFileLength);
		// 共通ヘッダ部にデータを格納する
		pSendFileIng->header.wCommand = htons(CMD_SEND_FILE_ING);
		// データ部にデータを格納する
		memcpy(pSendFileIng->bData, pbData, iReadSize);
	}
	else							// ファイルの最後をまで読んだので完了を送信
	{
		Locate(1, 10, 1);
		fprintf(stderr, "ファイル送信完了要求送信 %d %d", m_lTotalSentSize, iReadSize);
		// 共通ヘッダ部にデータを格納する
		pSendFileIng->header.wCommand = htons(CMD_SEND_FILE_END_REQ);
		fEnd = TRUE;				// ファイル送信完了
	}
	// パケット全体を送信バッファに書きこむ
	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileIng, iPacketSize);
	if (fRet == TRUE)				// パケットを書き込めた
	{
		if (fEnd == TRUE)			// 最後まで送信したので終わり
			iRet = 2;
		else
		{
			m_lTotalSentSize += iReadSize;	// パケットを書き込めたので読み込みポインタ更新
			iRet = 1;
		}
	}
	else							// 書き込めなかったので読み込みポインタはそのまま
	{
		Locate(1, 11, 0);
		fprintf(stderr, "送信バッファの空き待ち(%d)", ++m_iWaitCount);
		// 送信バッファに書けなかったのでちょっと待つ
		usleep(100 * 1000);			// READ_SIZEを送信するのに掛かる時間程度待つのが良いでしょう
	}
	SAFE_FREE(pSendFileIng)
	SAFE_FREE(pbData)
	return(iRet);
}


//----------------------------------------------
// function
//		送信中断
//		遷移状態変更、中断パケット送信
// parameter
//		なし
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::SendFileAbort()
{
	BOOL		fRet = FALSE;

	// ファイル送信中の時だけ中断
	if (m_iSendFileStatus != SEND_FILE_STAT_ING)
		goto L_END;

	m_iSendFileStatus = SEND_FILE_STAT_ABORT;
	m_dwStart = timeGetTime();
	fRet = SendSendFileAbortPacket();
L_END:
	return(fRet);
}

//----------------------------------------------
// function
//		★送信中断コマンドの送信
// parameter
//		なし
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::SendSendFileAbortPacket()
{
	BOOL		fRet = FALSE;
	SendFileAbortReqRec	*pSendFileAbortReq = NULL;
	int			iPacketSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(SendFileAbortReqRec);
	pSendFileAbortReq = (SendFileAbortReqRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pSendFileAbortReq->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileAbortReq->header.wCommand = htons(CMD_SEND_FILE_ABORT_REQ);
	pSendFileAbortReq->header.wDataLen = htons(0);

	Locate(1, 10, 1);
	fprintf(stderr, "ファイル送信中断要求送信");
	
	// ファイル送信中断パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileAbortReq, iPacketSize);
	SAFE_FREE(pSendFileAbortReq);
	return(fRet);
}

//----------------------------------------------
// function
//		★ファイル送信中断要求応答のチェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 2:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int CSendFileThread::CheckSendFileAbortRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cSendFileAbortRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cSendFileAbortRes == 1)			// 成功応答
		iRet = 2;
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		★ファイル送信コマンド/応答の解析
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗
//----------------------------------------------
int CSendFileThread::AnalyzeCmd()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	int			iSize = m_pCRingBuffCmd->GetReadableSize();
	int			iPacketSize, iDataSize;
	LPBYTE		pbPacket = NULL;

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする(パケット丸ごと格納されているはず)
	m_pCRingBuffCmd->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	wCmd = ntohs(Header.wCommand);
	iDataSize = ntohs(Header.wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);
	if (iSize < iPacketSize)				// パケットサイズに満たないときは何もしない
		goto L_END;

	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を読む
	m_pCRingBuffCmd->Read(pbPacket, iPacketSize);

	switch (wCmd)
	{
		case CMD_SEND_FILE_START_RES:			// 送信開始要求応答受信
			iRet = RecvSendFileStartResPacket((SendFileResRec *)pbPacket);
			break;
		case CMD_SEND_FILE_END_RES:				// 送信完了要求応答受信
			iRet = RecvSendFileEndResPacket((SendFileResRec *)pbPacket);
			break;
		case CMD_SEND_FILE_ABORT_RES:			// 送信中断要求応答受信
			iRet = RecvSendFileAbortResPacket((SendFileResRec *)pbPacket);
			break;
		default:
			iRet = -1;
			break;
	}
L_END:
	SAFE_FREE(pbPacket)
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信開始要求応答処理
// parameter
//		SendFileResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::RecvSendFileStartResPacket(SendFileResRec *pbPacket)
{
	int			iRet = 1;

	m_cSendFileStartRes = pbPacket->cResCode;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信完了要求応答処理
// parameter
//		SendFileResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::RecvSendFileEndResPacket(SendFileResRec *pbPacket)
{
	int			iRet = 1;

	m_cSendFileEndRes = pbPacket->cResCode;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信中断要求応答処理
// parameter
//		SendFileResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::RecvSendFileAbortResPacket(SendFileResRec *pbPacket)
{
	int			iRet = 1;

	m_cSendFileAbortRes = pbPacket->cResCode;
	return(iRet);
}

//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//	  TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::IsZombie()
{
	BOOL	fRet;

	m_pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pCMySyncObject->UnLock();
	return(fRet);
}

