#pragma once
#include	"stdThread.h"

#define		MAGIC_DATA_LEN		36
#define		MAGIC_STRING		"E2E6F7BF-42B1-4382-AF0B-1F452ED13EB6"
#define		FILE_NAME_LEN		256			// ファイル名の長さ
#define		FILE_SIZE_DIGIT		10			// ファイルサイズの桁数

#define		TIMEOUT_FOR_RES		(100000)		// 応答までのタイムアウト時間msec
//-----------------------------------------------------------------------
// チャットメッセージコマンド
// Client->Server, Server->Client
#define		CMD_MSG_DATA			100
//-----------------------------------------------------------------------
// ファイル送信コマンド 要求/応答
// Client->Server
#define		CMD_SEND_FILE_START_REQ	200		// 送信開始要求
#define		CMD_SEND_FILE_ING		201		// ファイル送信中
#define		CMD_SEND_FILE_END_REQ	202		// 送信完了要求
#define		CMD_SEND_FILE_ABORT_REQ	203		// 中断要求

// Server->Client
#define		CMD_SEND_FILE_START_RES	1200	// 送信開始要求に対する応答
#define		CMD_SEND_FILE_END_RES	1202	// 送信完了要求に対する応答
#define		CMD_SEND_FILE_ABORT_RES	1203	// 中断要求に対する応答
//-----------------------------------------------------------------------
// ファイル受信コマンド 要求/応答
// Client->Server
#define		CMD_RECV_FILE_SIZE_REQ	300		// 受信サイズ要求
#define		CMD_RECV_FILE_START_REQ	301		// 受信開始要求
#define		CMD_RECV_FILE_END_RES	1303	// 送信完了要求に対する応答
#define		CMD_RECV_FILE_ABORT_REQ	304		// 中断要求

// Server->Client
#define		CMD_RECV_FILE_SIZE_RES	1300	// 受信サイズ要求に対する応答
#define		CMD_RECV_FILE_ABORT_RES	1304	// 中断要求に対する応答
#define		CMD_RECV_FILE_ING		302		// ファイル受信中
#define		CMD_RECV_FILE_END_REQ	303		// 受信完了要求
//-----------------------------------------------------------------------
// ★データ転送方法 要求/応答
// Client->Server
#define		CTRL_TRANS_METHOD_REQ	400		// ★データ送受信方法要求(TCP/UDP)
// Server->Client
#define		CTRL_TRANS_METHOD_RES	1400	// ★データ送受信方法応答(TCP/UDP)
//-----------------------------------------------------------------------

#pragma pack(push ,1)
// 共通ヘッダ部
typedef struct
{
	BYTE	bMagicData[MAGIC_DATA_LEN];		// 不正アクセス防止用(WindowsのGUIDを使います)
	WORD	wCommand;						// ネットワークオーダー
	WORD	wDataLen;						// ネットワークオーダー
} HeaderRec;

//------------------------------------
// 共通ヘッダ部 + データ部 
//	メッセージ送信用構造体 
typedef struct
{
	HeaderRec	header;
	BYTE		bMsgData[1];				// nullターミネイトを含まないデータ
} MsgDataRec;

//------------------------------------
// データ部 
//	ファイル送信開始要求
//	ファイル受信サイズ要求		サイズ項目0で送信
//	ファイル受信サイズ要求応答	OKの時はファイルサイズに値が入る
//	ファイル受信開始要求			サイズ項目0のとき受信を実施しない
typedef struct
{
	char	szFileSize[FILE_SIZE_DIGIT + 1];	// ファイル長文字列(nullターミネートを含む)
	char	szFileName[FILE_NAME_LEN + 1];		// ファイル名文字列(nullターミネートを含む)UTF-8
} FileInfoRec;

// 共通ヘッダ部 + データ部 
typedef struct
{
	HeaderRec	header;
	FileInfoRec	FileInfo;
} SendFileStartReqRec, RecvFileSizeReqRec, RecvFileStartReqRec, RecvFileSizeResRec;

//------------------------------------
// データ部 
//	ファイル送信(ファイルの途中、最後)
//	ファイル受信(ファイルの途中、最後)
typedef struct
{
	char	szOffset[FILE_SIZE_DIGIT + 1];		// 先頭からの位置
	char	szSize[FILE_SIZE_DIGIT + 1];		// データのサイズ
} SendFileStatRec;

// 共通ヘッダ部 + データ部 
typedef struct
{
	HeaderRec		header;
	SendFileStatRec	Stat;
	BYTE			bData[1];
} SendFileIngRec, SendFileEndReqRec, RecvFileIngRec, RecvFileEndReqRec;

//------------------------------------
// 共通ヘッダ部 + データ部 
//	ファイル送信中断要求
//	ファイル受信中断要求
typedef struct
{
	HeaderRec	header;
} SendFileAbortReqRec, RecvFileAbortReqRec;

//------------------------------------
// 共通ヘッダ部 + データ部 
//	ファイル送信要求に対する応答
//	ファイル受信要求に対する応答
typedef struct
{
	HeaderRec	header;
	char		cResCode;						// 1:成功 -1:失敗 など応答内容コード
} SendFileResRec, RecvFileResRec;


//------------------------------------
// ★データ転送方法 要求応答
// データ部 
typedef struct
{
	BYTE	bMethod;							// 0:TCP 1:UDP
	WORD	wUDPSrcPort;						// 要求時クライアント送信元ポート番号
												// 応答時サーバ送信元ポート番号
} TransMethodInfoRec;

// 共通ヘッダ部 + データ部 
// ClientからUDP要求を出した時、Serverから応答がTCPならTCPでデータ転送を行う
// Serverからの応答がUDPなら返ってきたポート番号をクライアントは送信先として使用する
typedef struct
{
	HeaderRec	header;
	TransMethodInfoRec	TansMethodInfo;
} TransMethodReqRec, TransMethodResRec;

#pragma pack(pop)

